#if ! defined( _SLOG )
#define _SLOG

#include <stdio.h>
#include "fbuf.h"

/* the high-level API for reading and writing SLOG files */

#define SLOG_PROFILENAME  "SLOG_Profile.txt"
#define SLOG_RECDEFFILE   "SLOG_RecDefs.dat"

/*   The macro defines the indexing function for SLOG_recdefs_t.defs[]      */
/*   Hence (ix, iy )th element is accessed as ~defs[ rd_idx( ix, iy ) ]     */
#define  RecDefs_Ncolumn   3
#define  rd_idx( ix, iy )  ( (ix) + RecDefs_Ncolumn * (iy) )

/*
     Define the basic data types used, SLOG_xxx,
     and their printing formats, fmt_xxx, used in fprintf().
*/
typedef  int                 SLOG_int   ;
#define  fmt_int             "%d"

typedef  unsigned char       SLOG_bit   ;
#define  fmt_bit             "%hu"
typedef  unsigned char       SLOG_uint8 ;
#define  fmt_ui8             "%hu"
typedef  unsigned short      SLOG_uint16;
#define  fmt_ui16            "%hu"
typedef  unsigned int        SLOG_uint32;
#define  fmt_ui32            "%u"
typedef  unsigned long long  SLOG_uint64;
#define  fmt_ui64            "%llu"
typedef  double              SLOG_time  ;
#define  fmt_time            "%f"
typedef  SLOG_uint64         SLOG_fptr  ;
#define  fmt_fptr            fmt_ui64

/*
     Define the components and their correspong print format 
     in the __internal__ Interval Record
*/
typedef  SLOG_uint16         SLOG_bytesize_t;
#define  fmt_sz_t            fmt_ui16
typedef  SLOG_uint16         SLOG_rectype_t;
#define  fmt_rtype_t         fmt_ui16
typedef  SLOG_uint16         SLOG_evttype_t;
#define  fmt_etype_t         fmt_ui16
typedef  SLOG_time           SLOG_starttime_t;
#define  fmt_stime_t         fmt_time
typedef  SLOG_time           SLOG_duration_t;
#define  fmt_dura_t          fmt_time
typedef  SLOG_bit            SLOG_bebit_t;
#define  fmt_bebit_t         fmt_bit
typedef  SLOG_bit            SLOG_bebits_t;
#define  fmt_bebits_t        fmt_bit
typedef  SLOG_uint16         SLOG_nodeID_t; 
#define  fmt_nodeID_t        fmt_ui16
typedef  SLOG_uint8          SLOG_cpuID_t; 
#define  fmt_cpuID_t         fmt_ui8
typedef  SLOG_uint32         SLOG_threadID_t;
#define  fmt_thID_t          fmt_ui32
typedef  SLOG_fptr           SLOG_where_t;
#define  fmt_where_t         fmt_fptr

typedef  SLOG_uint16         SLOG_N_assocs_t;
#define  fmt_Nassocs_t       fmt_ui16
typedef  SLOG_uint32         SLOG_assoc_t;
#define  fmt_assoc_t         fmt_ui32

typedef  SLOG_uint16         SLOG_N_args_t;
#define  fmt_Nargs_t         fmt_ui16
typedef  SLOG_uint32         SLOG_arg_t;
#define  fmt_arg_t           fmt_ui32

#define  SLOG_fptr_NULL           ( (SLOG_fptr) 0 )

/*  Define User acroym for user convenience  */
#define  SLOG                     SLOG_STREAM *
#define  SLOG_Irec                SLOG_intvlrec_t *

/*  memory size for some internal data structure  */
#define SLOG_BUFFER_UNIT   256      /* byte size for frame buffer           */
#define SLOG_NDirEntry     128      /* Number of entries in Frame Directory */
#define SLOG_MAX_TYPES     30       /* Number of types in SLOG_typesz       */

/*  High level API and internal subroutine return error code  */
#define SLOG_SUCCESS       1
#define SLOG_FAIL          -1
#define SLOG_EOF           -2

/* ***  GLOBAL DATA TYPES  *** */

/* File descriptor for the output stream */
#define  outfile  stdout 
/* File descriptor for the error stream */
#define  errfile  stderr

enum SLOG_DataType { 
     bit, ui8, ui16, ui32, ui64, time, fptr,
     sz_t, rtype_t, etype_t, stime_t, dura_t,
     bebit_t, bebits_t, nodeID_t, cpuID_t, thID_t, where_t, 
     Nassocs_t, assoc_t, Nargs_t, arg_t,
     FrameDirHdr, FrameDirEntry, FrameHdr, min_IntvlRec
};

/*  Initialized in SLOG_InitGlobalData()  */
SLOG_uint32 SLOG_typesz[ SLOG_MAX_TYPES ];

struct {
    SLOG_rectype_t   RecDefs_MinIdx4VarIrec;
    SLOG_rectype_t   RecDefs_MaxIdx4VarIrec;
} SLOG_global;


/*   The def'n of the Fixed Record definition   */
typedef struct {
    SLOG_uint32       NfixedRec;
    SLOG_uint32      *defs;                     /*  defs[ NfixedRec * 3 ]  */
} SLOG_recdef_t;


/*   The def'n of the Entry of Frame Directory   */
typedef struct {
    SLOG_fptr    fptr2framehdr;
    SLOG_time    starttime;
    SLOG_time    endtime;
} SLOG_dir_entry_t;


/*   The def'n of the Frame Directory   */
typedef struct {
    SLOG_fptr          prevdir;
    SLOG_fptr          nextdir;
    SLOG_uint32        Nframe;
    SLOG_dir_entry_t  *entries;           /*  entries[ SLOG_NDirEntry ]  */
} SLOG_dir_t;


/*   The def'n of the Frame Header   */
typedef struct {
    SLOG_uint32   Nbytes_inc;
    SLOG_uint32   Nrec_inc;
    SLOG_uint32   Nbytes_cur;
    SLOG_uint32   Nrec_cur;
    SLOG_uint32   Nbytes_out;
    SLOG_uint32   Nrec_out;
} SLOG_framehdr_t;

/*   The def'n of an association pair, this is used to support association   */
typedef struct {
    /*
    SLOG_intvlrec_t   *irec;
    */
    void              *irec;             /* pointer to the associated irec */
    SLOG_assoc_t       fptr;             /* file pointer of associated irec */
} SLOG_assocpair_t;

/*   The def'n of a MPI call arguments vector   */
typedef struct {
    SLOG_N_args_t     size;
    SLOG_arg_t       *values;            /*  values[ size ]  */
} SLOG_vtrarg_t;

typedef struct SLOG_vtrarg_lptr_ {
           SLOG_vtrarg_t     *vtr;
    struct SLOG_vtrarg_lptr_ *prev;
    struct SLOG_vtrarg_lptr_ *next;
} SLOG_vtrarg_lptr_t;

/*   The def'n of the Interval Record ( run time )   */
typedef struct {
    SLOG_bytesize_t      bytesize;          /* bytesize in file */
    SLOG_rectype_t       rectype;
    SLOG_evttype_t       evttype;
    SLOG_bebits_t        bebits[2];
    SLOG_starttime_t     starttime;
    SLOG_duration_t      duration;
    SLOG_nodeID_t        node_id;
    SLOG_cpuID_t         cpu_id;
    SLOG_threadID_t      thread_id;
    SLOG_where_t         where;
    SLOG_N_assocs_t      N_assocs;
    SLOG_assoc_t        *assocs;             /* assocs[ N_assocs ] */
    SLOG_vtrarg_lptr_t  *vhead;
    SLOG_vtrarg_lptr_t  *vtail;
    SLOG_vtrarg_lptr_t  *vptr;
} SLOG_intvlrec_t;


/*   The def'n of Interval Record Bi-directional pointer type  */
typedef struct SLOG_intvlrec_lptr_ {
           SLOG_intvlrec_t       *irec;
    struct SLOG_intvlrec_lptr_   *prev;
    struct SLOG_intvlrec_lptr_   *next;
} SLOG_intvlrec_lptr_t;


/*   The def'n of Interval Record Bi-directional Linked List   */
typedef struct {
    SLOG_uint32            Nbytes_in_file;
    SLOG_uint32            count_irec;
    SLOG_intvlrec_lptr_t  *lhead;
    SLOG_intvlrec_lptr_t  *ltail;
    SLOG_intvlrec_lptr_t  *lptr;
} SLOG_intvlrec_blist_t;


/*   The def'n of the SLOG_STREAM   */
typedef struct {
    FILE          *fd;            /* File descriptor of the slogfile      */
    filebuf_t     *fbuf;          /* File Buffer pointer of the slogfile  */
    SLOG_fptr      file_dir_hdr;  /* beginning FD pointer in slog file    */
    SLOG_fptr      file_dir_cur;  /* current entry in FD in slog file     */
    SLOG_fptr      file_loc;      /* Current position in the slog file    */
    SLOG_recdef_t  rec_defs;      /* Interval Records Descriptor          */
    SLOG_dir_t              frame_dir;
    SLOG_dir_entry_t        frame_dir_entry;
    SLOG_framehdr_t         frame_hdr;
    SLOG_intvlrec_blist_t  *inc_bbuf;
    SLOG_intvlrec_blist_t  *cur_bbuf;
    SLOG_intvlrec_blist_t  *out_bbuf;
    SLOG_intvlrec_blist_t  *tmp_bbuf;
} SLOG_STREAM;



/*  High level SLOG (Common) API for both __Read__ and __Write__  */

SLOG_intvlrec_t *SLOG_Irec_Create( void );

void SLOG_Irec_Free( SLOG_intvlrec_t *irec );

void SLOG_Irec_CopyVal( const SLOG_intvlrec_t *src, SLOG_intvlrec_t *dest );

SLOG_intvlrec_t *SLOG_Irec_Copy( const SLOG_intvlrec_t *src );

int SLOG_Irec_Print( const SLOG_intvlrec_t *irec, FILE *outfd );

int SLOG_Irec_IsEmpty( const SLOG_intvlrec_t *irec );

int SLOG_Irec_IsConsistent( const SLOG_intvlrec_t *irec );

	/*  Low level SLOG_Irec Methods  */

int SLOG_Irec_DepositToFbuf( const SLOG_intvlrec_t *irec, filebuf_t *fbuf );

SLOG_intvlrec_t *SLOG_Irec_WithdrawFromFbuf( const SLOG_recdef_t  slog_recdefs,
                                                   filebuf_t     *fbuf );

int SLOG_Irec_AddVtrArgs(       SLOG_intvlrec_t *irec,
                          const SLOG_vtrarg_t   *varg );

const SLOG_vtrarg_t *SLOG_Irec_GetVtrArgs( SLOG_intvlrec_t  *irec );

void SLOG_Irec_DelAllVtrArgs( SLOG_intvlrec_t  *irec );



/*  High level SLOG __Write__ API  */

SLOG_STREAM *SLOG_OpenOutputStream( const char *filename );

int SLOG_Irec_SetMinRec(       SLOG_intvlrec_t  *intvlrec,
                         const SLOG_rectype_t    rectype,
                         const SLOG_evttype_t    evttype,
                         const SLOG_bebit_t      bebit_0,
                         const SLOG_bebit_t      bebit_1,
                         const SLOG_starttime_t  starttime,
                         const SLOG_duration_t   duration,
                         const SLOG_nodeID_t     node_id,
                         const SLOG_cpuID_t      cpu_id,
                         const SLOG_threadID_t   thread_id,
                         const SLOG_where_t      where );

int SLOG_Irec_SetAssocs(       SLOG_intvlrec_t  *intvlrec,
                         const SLOG_STREAM      *slog,
                         const SLOG_N_assocs_t   N_assocs,
                         const SLOG_assoc_t     *assocs  );

int SLOG_Irec_SetArgs(       SLOG_intvlrec_t  *intvlrec,
                       const SLOG_STREAM      *slog,
                       const SLOG_N_args_t     N_args, 
                       const SLOG_arg_t       *args );

int SLOG_Irec_ToOutputStream(       SLOG_STREAM      *slog,
                              const SLOG_intvlrec_t  *irec );

int SLOG_CloseOutputStream( SLOG_STREAM   *slog );



/*  High level SLOG __Read__ API  */

SLOG_STREAM *SLOG_OpenInputStream( const char *filename );

const SLOG_intvlrec_t *SLOG_Irec_FromInputStream( SLOG_STREAM *slog,
                                                  int *irc );

int SLOG_Irec_GetMinRec( const SLOG_intvlrec_t   *intvlrec,
                               SLOG_rectype_t    *rectype,
                               SLOG_evttype_t    *evttype,
                               SLOG_bebit_t      *bebit_0,
                               SLOG_bebit_t      *bebit_1,
                               SLOG_starttime_t  *starttime,
                               SLOG_duration_t   *duration,
                               SLOG_nodeID_t     *node_id,
                               SLOG_cpuID_t      *cpu_id,
                               SLOG_threadID_t   *thread_id,
                               SLOG_where_t      *where );

int SLOG_Irec_GetAssocs( const SLOG_intvlrec_t  *intvlrec,
                         const SLOG_N_assocs_t   size_assocs,
                               SLOG_assoc_t     *assocs );

int SLOG_Irec_GetArgs(       SLOG_intvlrec_t  *intvlrec,
                       const SLOG_N_args_t     size_args,
                             SLOG_arg_t       *args );

int SLOG_CloseInputStream( SLOG_STREAM *slog );

#endif  /* if ! defined( _SLOG ) */
